//
//  MembershipListTableViewController.swift
//  SampleApp
//
//  Created by Federico Zanetello on 4/10/19.
//  Copyright © 2019 David Zhang. All rights reserved.
//

import AmitySDK
import UIKit
import SwiftUI

final class MembershipListTableViewController: UITableViewController, DataSourceListener {
    weak var client: AmityClient!
    var channelId: String!
    var subChannelId: String!
    
    private var resultSearchController = UISearchController()
    private var dataSource: MembershipListDataSource!
    
    var debouncer = Debouncer(delay: 0.5)
    
    // MARK: Lifecycle
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        dataSource = MembershipListDataSource(client: client, channelId: channelId, subChannelId: subChannelId)
        dataSource.dataSourceObserver = self
        
        let nib = UINib(nibName: SampleAppTableViewCell.identifier, bundle: nil)
        tableView.register(nib, forCellReuseIdentifier: SampleAppTableViewCell.identifier)
        
        self.title = "Memberships"
        
        // remove empty rows
        tableView.tableFooterView = UIView()
        
        resultSearchController = ({
            let controller = UISearchController(searchResultsController: nil)
            controller.searchResultsUpdater = self
            controller.searchBar.sizeToFit()
            
            controller.searchBar.placeholder = "Search Roles (seperate by ',')"
            tableView.tableHeaderView = controller.searchBar
            return controller
        })()
        
        let permissionButton = UIBarButtonItem(image: UIImage(systemName: "arrow.up.right.circle"), style: .plain, target: self, action: #selector(onPermissionButtonTap))
        let filterButton = UIBarButtonItem(image: UIImage(systemName: "line.horizontal.3.decrease.circle"), style: .plain, target: self, action: #selector(onFilterButtonTap))
        let addMemberButton = UIBarButtonItem(image: UIImage(systemName: "person.fill"), style: .plain, target: self, action: #selector(onAddMemberButtonTap))
        
        
        self.navigationItem.rightBarButtonItems = [permissionButton, filterButton, addMemberButton]
        
        self.dataSource.fetchListOfMembers()
    }
    
    @objc func onPermissionButtonTap() {
        let permissionController = UIHostingController(rootView: UserPermissionView(channelId: channelId))
        permissionController.title = "Channel Permission"
        self.navigationController?.pushViewController(permissionController, animated: true)
    }
    
    @objc func onFilterButtonTap() {
        let actionSheet = UIAlertController(title: "Apply Sort & Filter", message: "", preferredStyle: .actionSheet)
        actionSheet.addAction(UIAlertAction(title: "All Members", style: .default, handler: { [weak self] action in
            self?.title = "All Members"
            self?.dataSource.updateMembershipFilter(filter: .all)
        }))
        
        actionSheet.addAction(UIAlertAction(title: "Banned Members", style: .default, handler: { [weak self] action in
            self?.title = "Banned Members"
            self?.dataSource.updateMembershipFilter(filter: .ban)
        }))
        
        actionSheet.addAction(UIAlertAction(title: "Muted Members", style: .default, handler: { [weak self] action in
            self?.title = "Muted Members"
            self?.dataSource.updateMembershipFilter(filter: .mute)
        }))
        
        actionSheet.addAction(UIAlertAction(title: "Sort: Last Created", style: .default, handler: { [weak self] action in
            self?.dataSource.updateMembershipSort(sortBy: .lastCreated)
        }))
        
        actionSheet.addAction(UIAlertAction(title: "Sort: First Created", style: .default, handler: { [weak self] action in
            self?.dataSource.updateMembershipSort(sortBy: .firstCreated)
        }))
        
        actionSheet.addAction(UIAlertAction(title: "Dismiss", style: .cancel, handler: nil))
        self.present(actionSheet, animated: true, completion: nil)
    }
    
    @objc func onAddMemberButtonTap() {
        showPopupWithTextField(title: "Add Member", message: "Add member to this channel", placeholder: "Enter user id") { userId in
            
            self.dataSource.addMember(id: userId) { [weak self] success, error in
                if let err = error {
                    self?.showAlert(message: "Error while adding member: \(err)")
                }
            }
        }
    }
    
    // MARK: UITableViewDataSource
    
    override func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return dataSource.memberListCount
    }

    override func tableView(_ tableView: UITableView, willDisplay cell: UITableViewCell, forRowAt indexPath: IndexPath) {
        if indexPath.row == dataSource.memberListCount - 2 {
            dataSource.fetchNextPage()
        }
    }
    
    override func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        guard let cell = tableView.dequeueReusableCell(withIdentifier: SampleAppTableViewCell.identifier) as? SampleAppTableViewCell else { return UITableViewCell() }
        
        if let membershipData = dataSource.getMember(at: indexPath) {
            cell.titleLabel.text = membershipData.userId
            setTags(in: cell, for: membershipData)
        }
        
        return cell
    }
    
    override func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        showAvailableActionsForMember(at: indexPath)
    }
    
    private func setTags(in cell: SampleAppTableViewCell, for membership: AmityChannelMember) {
        let tags: [(text: String, color: UIColor)] = self.createTags(for: membership)
        
        let mutableAttributedString = NSMutableAttributedString()
        for tag in tags {
            let attributedTagString = NSAttributedString(string: " \(tag.text) ",
                                                         attributes: [.backgroundColor: tag.color,
                                                                      .foregroundColor: .white])
            mutableAttributedString.append(attributedTagString)
            mutableAttributedString.append(NSAttributedString(string: " "))
        }
        
        cell.subtitleLabel.attributedText = mutableAttributedString
    }
    
    private func createTags(for membership: AmityChannelMember) -> [(String, UIColor)] {
        let pastelColors: [UIColor] = [UIColor(named: "PastelRed"),
                                       UIColor(named: "PastelOrange"),
                                       UIColor(named: "PastelYellow"),
                                       UIColor(named: "PastelGreen"),
                                       UIColor(named: "PastelBlue"),
                                       UIColor(named: "PastelGreen"),
                                       UIColor(named: "NeonPink"),
                                       UIColor(named: "NeonRed"),
                                       UIColor(named: "NeonGreen"),
                                       UIColor(named: "NeonAzure")].compactMap { $0 }
        var tags: [(String, UIColor)] = []
        
        if membership.isBanned {
            tags.append(("Banned", pastelColors[tags.count % pastelColors.count]))
        }
        
        if membership.isMuted {
            tags.append(("Muted", pastelColors[tags.count % pastelColors.count]))
        }
        
        for role in membership.roles{
            tags.append((role, pastelColors[tags.count % pastelColors.count]))
        }
        
        return tags
    }
    
    // MARK: DataSourceListener
    
    func didUpdateDataSource() {
        Log.add(info: "Reloading member list data source")
        tableView.reloadData()
    }

    func showAvailableActionsForMember(at indexPath: IndexPath) {
        guard let member = dataSource.getMember(at: indexPath) else { return }
        
        let actionSheet = UIAlertController(title: "Choose Action", message: "", preferredStyle: .actionSheet)
        
        actionSheet.addAction(UIAlertAction(title: "Member Info", style: .default, handler: { [weak self] action in
            actionSheet.dismiss(animated: true) {
                self?.showMemberInfo(member: member)
            }
        }))
        
        actionSheet.addAction(UIAlertAction(title: "Add Role", style: .default, handler: { [weak self] action in
            self?.showAlert(type: .add, for: indexPath)
        }))
        
        actionSheet.addAction(UIAlertAction(title: "Remove Role", style: .default, handler: { [weak self] action in
            self?.showAlert(type: .remove, for: indexPath)
        }))
        
        let currentUserId = AmityManager.shared.client?.currentUserId ?? ""
        if currentUserId != member.userId {
            if member.isBanned {
                actionSheet.addAction(UIAlertAction(title: "Unban User", style: .default, handler: { [weak self] action in
                    self?.dataSource.unbanMember(at: indexPath, completion: { isSuccess, error in
                        let info = isSuccess ? "Unban Successful" :  "Unban Failed"
                        self?.showAlert(message: "\(info), Error: \(String(describing: error))")
                    })
                }))
            } else {
                actionSheet.addAction(UIAlertAction(title: "Ban User", style: .default, handler: { [weak self] action in
                    self?.dataSource.banMember(at: indexPath, completion: { isSuccess, error in
                        let info = isSuccess ? "Ban Successful" :  "Ban Failed"
                        self?.showAlert(message: "\(info), Error: \(String(describing: error))")
                    })
                }))
            }
        }
        
        actionSheet.addAction(UIAlertAction(title: "Remove Member", style: .default, handler: { [weak self] action in
            self?.dataSource.removeMember(id: member.userId, completion: { success, error in
                if let err = error {
                    self?.showAlert(message: "Error while removing member: \(err)")
                }
            })
        }))
        
        actionSheet.addAction(UIAlertAction(title: "Mute Member (60 Seconds)", style: .default, handler: { [weak self] action in
            self?.dataSource.muteMember(at: indexPath, completion: { success, error in
                if let err = error {
                    self?.showAlert(message: "Error while removing member: \(err)")
                }
            })
        }))
        
        actionSheet.addAction(UIAlertAction(title: "UnMute Member", style: .default, handler: { [weak self] action in
            self?.dataSource.unMuteMember(at: indexPath, completion: { success, error in
                if let err = error {
                    self?.showAlert(message: "Error while removing member: \(err)")
                }
            })
        }))
        
        actionSheet.addAction(UIAlertAction(title: "Dismiss", style: .cancel, handler: nil))
        
        self.present(actionSheet, animated: true, completion: nil)
    }
}

extension MembershipListTableViewController: UISearchResultsUpdating {
    
    func updateSearchResults(for searchController: UISearchController) {
        debouncer.setCallback { [weak self] in
            self?.dataSource.filterMembers(by: searchController.searchBar.text)
        }
        
        debouncer.call()
    }
}

//MARK:- Alert for role input
private extension MembershipListTableViewController {
    
    func showPopupWithTextField(title: String, message: String, placeholder: String, completion: @escaping (String) -> Void) {
        let alert = UIAlertController(title: "Add Member", message: message, preferredStyle: .alert)
        
        alert.addTextField { (textField) in
            textField.placeholder = placeholder
        }
        
        alert.addAction(UIAlertAction(title: "OK", style: .default, handler: { action in
            guard let textField = alert.textFields?.first, let text = textField.text, !text.isEmpty else { return }
            completion(text)
        }))
        
        alert.addAction(UIAlertAction(title: "Cancel", style: .cancel, handler: nil))
        
        self.present(alert, animated: true, completion: nil)
    }
    
    func showAlert(type actionType: RoleActionType, for indexPath: IndexPath) {
        let alert = UIAlertController(title: "Role action", message: actionType.message, preferredStyle: .alert)
        
        alert.addTextField { (textField) in
            textField.placeholder = "Write a role"
        }
        
        alert.addAction(UIAlertAction(title: "OK", style: .default, handler: { [weak self, weak alert] (_) in
            guard let textField = alert?.textFields?.first, let text = textField.text, !text.isEmpty else { return }
            switch actionType {
            case .add:
                self?.dataSource.addRole(for: indexPath, role: text)
            case .remove:
                self?.dataSource.removeRole(for: indexPath, role: text)
            }
        }))
        
        alert.addAction(UIAlertAction(title: "Cancel", style: .cancel, handler: nil))
        
        self.present(alert, animated: true, completion: nil)
    }
    
    private func showAlert(message: String) {
        let alertController = UIAlertController(title: nil, message: message, preferredStyle: .alert)
        let ok = UIAlertAction(title: "OK", style: .default, handler: nil)
        alertController.addAction(ok)
        present(alertController, animated: true, completion: nil)
    }
    
    private func showMemberInfo(member: AmityChannelMember) {
        let information = """
        User id: \(member.userId)
        Display name: \(member.displayName ?? "-")
        Joined date: \(member.createdAt)
        Updated date: \(member.updatedAt)
        """
        
        let alert = UIAlertController(title: nil, message: information, preferredStyle: .alert)
        alert.addAction(UIAlertAction(title: "OK", style: .default, handler: nil))
        self.present(alert, animated: true, completion: nil)
    }
    
}

enum RoleActionType: Int {
    case add
    case remove
    
    var message: String {
        switch self {
        case .add: return "Add a role to the member"
        case .remove: return "Remove a role form the member"
        }
    }
}
