//
//  HomeViewController.swift
//  SampleApp
//
//  Created by Federico Zanetello on 4/30/19.
//  Copyright © 2019 David Zhang. All rights reserved.
//

import UIKit
import AmitySDK
import UserNotifications
import Combine

final class HomeViewController: UIViewController, ChannelsTableViewControllerDelegate, ChannelCreator, AmityClientDelegate {
    
    @IBOutlet private weak var displayNameButton: UIButton!
    @IBOutlet private weak var totalMessageCountLabel: UILabel!
    
    private var connectionStatusObservationToken: NSKeyValueObservation?
    private var totalUnreadCountObservationToken: NSKeyValueObservation?
    
    // Only place where client should be held strongly
    var client: AmityClient = AmityManager.shared.client!
    lazy var channelRepository = AmityChannelRepository(client: client)
    lazy var fileRepository = AmityFileRepository(client: client)
    lazy var pushNotificationManager = PushNotificationRegistrationManager(client: client)
    var channelType: AmityChannelType = .community
    
    private var channelListViewController: ChannelListTableViewController!
    private var displayNameController: DisplayNameButtonController!
    private var unreadCountController: UnreadCountLabelController!
    
    var channelUpdateToken: AmityNotificationToken?
    var sessionStateObservation: AnyCancellable?
    
    // MARK: Lifecyle
    
    override func awakeFromNib() {
        super.awakeFromNib()
        registerForPushNotifications()
    }
    
    func registerForPushNotifications() {
        UNUserNotificationCenter.current()
            .requestAuthorization(options: [.alert, .sound, .badge]) { granted, _ in
                guard granted else { return }
                DispatchQueue.main.async {
                    UIApplication.shared.registerForRemoteNotifications()
                }
            }
    }
    
    func didReceiveError(error: Error) {
        Log.add(info: "🔥 Delegate: Erorr Received \(error)")
        let error = error as NSError
        guard let _ = AmityErrorCode(rawValue: error.code) else {
            assertionFailure("unknown error \(error.code), please report this code to Amity")
            return
        }
        
        UIHelper.displayAlert(in: self, title: "Amity Error", message: String(describing: error))
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        client.delegate = self
        
        let userId = UserDefaults.standard.userId ?? ""
        
        if AmityManager.shared.needsLogin {
            performUserLogin(userId: userId, displayName: nil)
        } else {
            fetchChannelInformation()
        }
        
        trackConnectionStatus()
        trackUnreadCount()
        
        displayNameController = DisplayNameButtonController(displayNameButton: displayNameButton)
        unreadCountController = UnreadCountLabelController(unreadCountLabel: totalMessageCountLabel,
                                                           repository: channelRepository)
    }
    
    func performUserLogin(userId: String, displayName: String?) {
        Log.add(info: "Logging In User: \(userId) DisplayName: \(String(describing: displayName))")
        UserDefaults.standard.set(userId, forKey: "userId")
        Task { @MainActor in
            do {
                let authToken: String?
                if AmityManager.shared.needsAuthToken {
                    authToken = try await AmityManager.shared.getAuthToken(userId: userId)
                } else {
                    authToken = nil
                }
                try await client.login(userId: userId, displayName: displayName, authToken: authToken, sessionHandler: MySessionHandler())
                fetchChannelInformation()
                
                if let token = UserDefaults.standard.deviceToken, UserDefaults.standard.isRegisterdForPushNotification {
                    let pushNotificationManager =  PushNotificationRegistrationManager(client: client)
                    pushNotificationManager.register(token: token) { (result) in
                        switch result {
                        case .success:
                            Log.add(info: "✅ Device is registered: \(token)")
                        case .failure(let error):
                            Log.add(info: "🛑 Failed to register: \(error.localizedDescription)")
                        }
                    }
                }
            } catch {
                UIHelper.displayAlert(in: self, title: "Login Error", message: error.localizedDescription)
            }
        }
    }
    
    func fetchChannelInformation() {
        channelListViewController.fetchChannelList()
    }
    
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        switch segue.destination {
        case let destination as ChannelListTableViewController:
            channelListViewController = destination
            channelListViewController.delegate = self
            channelListViewController.repository = channelRepository
            channelListViewController.channelType = channelType
        case is SessionStateLEDViewController:
            break
        case let destination as MoreTableViewController:
            destination.client = client
        case let destination as CreateChannelTableViewController:
            destination.channelCreator = self
        default:
            break
        }
    }
    
    private var disposeBag: Set<AnyCancellable> = []
    
    private func trackUnreadCount() {
        channelRepository.$totalDefaultSubChannelUnreadCount.sink { [weak self] unreadCount in
            self?.totalMessageCountLabel.text = "\(unreadCount)"
            
        }.store(in: &disposeBag)
    }
    
    private func trackConnectionStatus() {
        sessionStateObservation = client.$sessionState.sink { [weak self] sessionState in
            switch sessionState {
            case .established:
                DispatchQueue.main.asyncAfter(deadline: .now() + .milliseconds(100)) {
                    if let currentUser = AmityManager.shared.client?.currentUser {
                        self?.displayNameController.observe(userObject: currentUser)
                    }
                }
            default:
                break
            }
        }
    }
    
    @IBAction private func titleTapped() {
        let titleText = "Enter Your New Display Name"
        let alertController = UIAlertController(title: titleText, message: nil, preferredStyle: .alert)
        
        let renameAction = UIAlertAction(title: "Rename", style: .default) { _ in
            guard let newDisplayName: String = alertController.textFields?.first?.text else { return }
            
            UserUpdateManager.shared.updateDisplayName(displayName: newDisplayName, completion: nil)
        }
        alertController.addAction(renameAction)
        
        alertController.addTextField { (textField) in
            textField.placeholder = self.client.currentUser?.object?.displayName
        }
        
        let cancelAction = UIAlertAction(title: "Cancel", style: .cancel, handler: nil)
        alertController.addAction(cancelAction)
        
        present(alertController, animated: true, completion: nil)
    }
    
    @IBAction private func profileTap() {
        let alertController = UIAlertController(title: "New login", message: nil, preferredStyle: .alert)
        
        alertController.addTextField(configurationHandler: { textField in
            textField.placeholder = "UserId"
        })
        
        let loginAction = UIAlertAction(title: "Login", style: .default, handler: { _ in
            guard let userId = alertController.textFields?.first?.text else { return }
            
            // Unregister for push notifications
            self.pushNotificationManager.unregisterUser { _ in
            }
            self.pushNotificationManager.unregisterDevice { _ in
            }
            
            // Logs out current client
            self.client.logout()
            
            // Perform new login
            self.performUserLogin(userId: userId, displayName: nil)
        })
        alertController.addAction(loginAction)
        
        let cancelAction = UIAlertAction(title: "Cancel", style: .cancel, handler: nil)
        alertController.addAction(cancelAction)
        
        present(alertController, animated: true, completion: nil)
    }
    
    // MARK: ChannelsTableViewControllerDelegate
    
    func joinChannel(channel: AmityChannel) {
        let channelId = channel.channelId
        let client = self.client

        switch channel.channelType {
        case .conversation:
            // You cannot join or leave conversation channel
            displaySubChannelList(for: channel)
        default:
            Task { [weak self] in
                do {
                    let channel = try await channelRepository.joinChannel(channelId: channelId)
                    await MainActor.run {
                        self?.displaySubChannelList(for: channel)
                    }
                } catch let err as NSError where err.isAmityErrorCode(.userIsBanned)  {
                    await MainActor.run {
                        UIHelper.displayAlert(in: self!, title: "Error", message: "Error Joining Channel: \(err.localizedDescription)")
                    }
                }
            }
        }
    }
    
    private func displaySubChannelList(for channel: AmityChannel) {
        let subChannelListController = SubChannelListController()
        subChannelListController.channel = channel
        
        self.navigationController?.pushViewController(subChannelListController, animated: true)
    }
    
    // MARK: ChannelCreator
    
    func createChannel(type: AmityChannelType, userIds: [String], avatar: UIImage?) {
        let randomId = UUID().uuidString
        let builder: AmityChannelBuilder? = {
            switch type {
            case .live:
                Log.add(info: "Creating Live channel")
                let builder = AmityLiveChannelBuilder()
                builder.setUserIds(userIds)
                builder.setDisplayName("my-live-channel_\(randomId)")
                builder.setTags(["ch-live","ios-sdk"])
                builder.setMetadata(["sdk_type":"ios"])
                return builder
            case .community:
                Log.add(info: "Creating community channel")
                let builder = AmityCommunityChannelBuilder()
                builder.setUserIds(userIds)
                builder.setDisplayName("my-community-channel_\(randomId)")
                builder.setTags(["ch-comm","ios-sdk"])
                builder.setMetadata(["sdk_type":"ios"])
                return builder
            case .conversation:
                Log.add(info: "Creating conversation channel")
                let builder = AmityConversationChannelBuilder()
                builder.setUserIds(userIds)
                builder.setDisplayName("my-conversation-channel_\(randomId)")
                builder.setTags(["ch-conv","ios-sdk"])
                builder.setMetadata(["sdk_type":"ios"])
                return builder
            default:
                Log.add(info: "Private & Standard channel types are depreciated.")
                return nil
            }
        } ()
        
        guard let builder = builder else {
            print("Unable to create channel builder")
            return
        }
        
        Task { [weak self] in
            do {
                let channel = try await channelRepository.createChannel(with: builder)
                self?.joinChannel(channel: channel)
                
                // Just showing how to set avatar.
                // You can upload the image first and set avatar data when creating channel
                if let avatar = avatar {
                    self?.uploadChannelAvatar(avatar: avatar, channelId: channel.channelId)
                }
            } catch {
                let lines = [
                    "Create channel error (channel with same id may already exist)",
                    "error: \(error.localizedDescription)"
                ]
                
                UIHelper.displayAlert(in: self!, title: "Error", message: lines.joined(separator: "\n"))
            }
        }
    }
    
    func uploadChannelAvatar(avatar: UIImage, channelId: String) {
        Task { @MainActor in
            do {
                let result = try await fileRepository.uploadImage(avatar, progress: nil)
                let builder = AmityChannelUpdateBuilder(channelId: channelId)
                builder.setAvatar(result)
                let _ = try await self.channelRepository.editChannel(with: builder)
            } catch {
                UIHelper.displayAlert(in: self, title: "Error", message: error.localizedDescription)
            }
        }
    }
}

