//
//  UserPostsFeedDataSource.swift
//  SampleApp
//
//  Created by Nishan Niraula on 5/5/20.
//  Copyright © 2020 David Zhang. All rights reserved.
//

import AmitySDK
import UIKit

enum FeedType {
    case globalFeed
    case customPostRankingGlobalFeed
    case myFeed
    case userFeed
    case singlePost
    case community
}

class UserPostsFeedManager {
    
    let client: AmityClient
    let feedRepository: AmityFeedRepository
    let postRepository: AmityPostRepository
    let pollRepository: AmityPollRepository
    
    var postCollection: AmityCollection<AmityPost>?
    var feedCollectionToken: AmityNotificationToken?
    
    var communityMembership: AmityCommunityMembership?
    
    var editedPost: AmityPost?
    let userId: String?
    let userName: String?
    var community: CommunityListModel?
    var postId: String? // Post id for individual post
    var feedType: FeedType = .myFeed
    
    var reactionCollection: AmityCollection<AmityReaction>?
    var reactionCollectionToken: AmityNotificationToken?
    
    var postObjectToken: AmityNotificationToken?
    var individualPost: AmityPost?
    
    var imageCache = [String: UIImage?]()
    var fileRepository: AmityFileRepository?
    
    var reactionRepository: AmityReactionRepository
    
    var comCategoryCollection: AmityCollection<AmityCommunityCategory>?
    var comCategoryCollectionToken: AmityNotificationToken?
    
    let uploadTracker = DispatchGroup()
    
    var flagger = PostFlagManager()
    
    var feedSortOption: AmityPostQuerySortOption = .lastCreated
    var feedSortCommunityOption: FeedItemDefaultAction = .publishedAndSortLastCreated
    var includeDeletedPosts = true
    
    lazy var dateFormatter: DateFormatter = {
        let formatter = DateFormatter()
        formatter.dateFormat = "MMM d, HH:mm"
        return formatter
    }()
    
    var postFeedType: AmityFeedType = .published
    var sortCommunityHandler: (() -> Void)?
    
    // Make sure your AmityClient is setup and then setup AmityFeedRepository & AmityPostCreator
    init(client: AmityClient, userId: String?, userName: String? = nil, community: CommunityListModel? = nil) {
        self.client = client
        self.userId = userId
        self.userName = userName
        self.community = community
        self.feedRepository = AmityFeedRepository(client: client)
        self.postRepository = AmityPostRepository(client: client)
        self.pollRepository = AmityPollRepository(client: client)
        self.fileRepository = AmityFileRepository(client: client)
        self.reactionRepository = AmityReactionRepository(client: client)
    }
    
    // MARK:- Feed Observer
    
    // Listen to feed changes. Anytime a post is created, observer
    // would be notified.
    // Note: Retain that AmityNotificationToken
    func observePostsFeedChanges(changeHandler:@escaping ()->()) {
                
        switch feedType {
            
        case .globalFeed:
            postCollection = feedRepository.getGlobalFeed()
            feedCollectionToken?.invalidate()
            feedCollectionToken = postCollection?.observe({(collection, _, error) in
                Log.add(info: "Post collection observer called")
                changeHandler()
            })
        case .customPostRankingGlobalFeed:
            postCollection = feedRepository.getCustomRankingGlobalFeed()
            feedCollectionToken?.invalidate()
            feedCollectionToken = postCollection?.observe({(collection, _, error) in
                Log.add(info: "Post collection observer called")
                changeHandler()
            })
        case .myFeed, .userFeed, .community:
            sortCurrentFeed(option: feedSortOption)
            
            feedCollectionToken?.invalidate()
            feedCollectionToken = postCollection?.observe({(collection, _, error) in
                Log.add(info: "Post collection observer called")
                changeHandler()
            })
        case .singlePost:
            guard let id = postId else {
                Log.add(info: "Error: Post id is nil while fetching individual post")
                return
            }
            
            let postObject = postRepository.getPost(withId: id)
            postObjectToken?.invalidate()
            postObjectToken = postObject.observe({ [weak self] (object, error) in
                self?.individualPost = object.object
                
                changeHandler()
            })
        }
    }
    
    func getPostAtIndex(index: Int) -> AmityPost? {
        switch feedType {
        case .globalFeed, .customPostRankingGlobalFeed, .myFeed, .userFeed, .community:
            return postCollection?.object(at: index)
        case .singlePost:
            return individualPost
        }
    }

    // MARK:- Reaction Observers
    
    func observeReactionsForPost(post: AmityPost?, reaction: String, completion: @escaping ()->()) {
        guard let post = post else { return }
        
        reactionCollectionToken?.invalidate()
        
        reactionCollection = reactionRepository.getReactions(post.postId, referenceType: .post, reactionName: reaction)
        reactionCollectionToken = reactionCollection?.observe({ (collection, change, error) in
            completion()
        })
    }
    
    func observeAllReactionsForPost(post: AmityPost?, completion: @escaping ()->()) {
        guard let post = post else { return }
        
        reactionCollection = reactionRepository.getReactions(post.postId, referenceType: .post, reactionName: nil)
        reactionCollectionToken = reactionCollection?.observe({ (collection, _, error) in
            completion()
        })
    }
    
    func getReactionAtIndex(index: Int) -> AmityReaction? {
        return reactionCollection?.object(at: index)
    }
    
    // Creates the post
    func createPost(text: String?,
                    images: [AmityImageData]?,
                    videos: [AmityVideoData]?,
                    files: [AmityFileData]?,
                    communityId: String?,
                    streamId: String?,
                    metadata: [String: Any]?,
                    mentionees: AmityMentioneesBuilder?,
                    completion: @escaping (Result<AmityPost, Error>) -> ()) {
        
        let targetId: String? = communityId == nil ? userId : communityId
        let targetType: AmityPostTargetType = communityId == nil ? .user : .community
        
        if images?.isEmpty == false && videos?.isEmpty == false {
            print("This app does not support create post with both video and image attached.")
            completion(.failure(GeneralError.unknown))
        }
        
        if let streamId = streamId, !streamId.isEmpty {
            Task { @MainActor in
                do {
                    let builder = AmityLiveStreamPostBuilder(streamId: streamId, text: text)
                    let post = try await postRepository.createLiveStreamPost(builder, targetId: targetId, targetType: targetType, metadata: metadata, mentionees: mentionees)
                    completion(.success(post))
                } catch {
                    completion(.failure(error))
                }
            }
        } else if let filesData = files, !filesData.isEmpty {
            
            // File Post
            
            Log.add(info: "--- Creating File Post ---")
            
            createFilePost(fileIds: filesData, text: text, targetId: targetId, targetType: targetType, metadata: metadata, mentionees: mentionees, completion: completion)
            
        } else if let imagesData = images, !imagesData.isEmpty {
            
            // Image Post
            
            Log.add(info: "--- Creating Image Post ---")
            
            createImagePost(imagesData: imagesData, text: text, targetId: targetId, targetType: targetType, metadata: metadata, mentionees: mentionees, completion: completion)
            
        } else if let videosData = videos, !videosData.isEmpty {
            
            // Video Post
            
            Log.add(info: "--- Creating Video Post ---")
            Log.add(info: "Uploading \(videosData.count) videos...")
            
            createVideoPost(videosData: videosData, text: text, targetId: targetId, targetType: targetType, metadata: metadata, mentionees: mentionees, completion: completion)
            
        } else {
            // Text Post
            Task { @MainActor in
                do {
                    let textBuilder = AmityTextPostBuilder()
                    if let text = text {
                        textBuilder.setText(text)
                    }
                    let post = try await postRepository.createTextPost(textBuilder, targetId: targetId, targetType: targetType, metadata: metadata, mentionees: mentionees)
                    completion(.success(post))
                } catch {
                    completion(.failure(error))
                }
            }
        }
    }
    
    func createFilePost(fileIds: [AmityFileData], text: String?, targetId: String?, targetType: AmityPostTargetType, metadata: [String: Any]?, mentionees: AmityMentioneesBuilder?, completion: @escaping (Result<AmityPost, Error>) -> ()) {
        Task { @MainActor in
            do {
                let filePostBuilder = AmityFilePostBuilder()
                if let text = text {
                    filePostBuilder.setText(text)
                }
                filePostBuilder.setFiles(fileIds)
                let post = try await postRepository.createFilePost(filePostBuilder, targetId: targetId, targetType: targetType, metadata: metadata, mentionees: mentionees)
                completion(.success(post))
            } catch {
                completion(.failure(error))
            }
        }
    }
    
    func createImagePost(imagesData: [AmityImageData], text: String?, targetId: String?, targetType: AmityPostTargetType,metadata: [String: Any]?, mentionees: AmityMentioneesBuilder?, completion: @escaping (Result<AmityPost, Error>) -> ()) {
        Task { @MainActor in
            do {
                let imagePostBuilder = AmityImagePostBuilder()
                if let text = text {
                    imagePostBuilder.setText(text)
                }
                imagePostBuilder.setImages(imagesData)
                let post = try await postRepository.createImagePost(imagePostBuilder, targetId: targetId, targetType: targetType, metadata: metadata, mentionees: mentionees)
                completion(.success(post))
            } catch {
                completion(.failure(error))
            }
        }
    }
    
    func createVideoPost(videosData: [AmityVideoData], text: String?, targetId: String?, targetType: AmityPostTargetType, metadata: [String: Any]?, mentionees: AmityMentioneesBuilder?, completion: @escaping (Result<AmityPost, Error>) -> ()) {
        Task { @MainActor in
            do {
                let videoPostBuilder = AmityVideoPostBuilder()
                if let text = text {
                    videoPostBuilder.setText(text)
                }
                videoPostBuilder.setVideos(videosData)
                let post = try await postRepository.createVideoPost(videoPostBuilder, targetId: targetId, targetType: targetType, metadata: metadata, mentionees: mentionees)
                completion(.success(post))
            } catch {
                completion(.failure(error))
            }
        }
    }
    
    // MARK: - Poll
    func createPollPost(text: String?,
                        numOptions: String?,
                        numDayToClose: String?,
                        isMultipleVote: Bool,
                        communityId: String?,
                        metadata: [String: Any]?,
                        mentionees: AmityMentioneesBuilder?,
                        completion: @escaping (Result<AmityPost, Error>) -> ()) {
        guard let text = text, let options = Int(numOptions ?? "2"), (options > 1 && options <= 10) else {
            completion(.failure(GeneralError.unknown))
            return
        }
        
        let targetId: String? = communityId == nil ? userId : communityId
        let targetType: AmityPostTargetType = communityId == nil ? .user : .community
        
        let pollBuilder = AmityPollCreateOptions()
        
        pollBuilder.setQuestion(text)
        if let numDayToClose = numDayToClose, let numDay = Int(numDayToClose) {
            let milliSec = numDay * (24 * 60 * 60 * 1000)
            pollBuilder.setTimeToClosePoll(milliSec)
        }
        pollBuilder.setAnswerType(isMultipleVote ? .multiple : .single)
        
        for index in 1...options {
            pollBuilder.setAnswer("Poll Option \(index)")
        }
        
        Task { @MainActor in
            do {
                let pollId = try await pollRepository.createPoll(pollBuilder)
                let builder = AmityPollPostBuilder()
                builder.setText(text)
                builder.setPollId(pollId)
                let post = try await postRepository.createPollPost(builder, targetId: targetId, targetType: targetType, metadata: metadata, mentionees: mentionees)
                completion(.success(post))
            } catch {
                completion(.failure(error))
            }
        }
    }
    
    func votePoll(withPollid pollId: String, answerIds: [String], onCompletion: @escaping (_ isSuccess: Bool)->()) {
        Task { @MainActor in
            do {
                let result = try await pollRepository.votePoll(withId: pollId, answerIds: answerIds)
                onCompletion(result)
            } catch {}
        }
    }
    
    func closedPoll(withPollId pollId: String, onCompletion: @escaping (_ isSuccess: Bool)->() ) {
        Task { @MainActor in
            do {
                let result = try await pollRepository.closePoll(withId: pollId)
                onCompletion(result)
            } catch {}
        }
    }
    
    func deletePoll(withPollId pollId: String, onCompletion: @escaping (_ isSuccess: Bool)->() ) {
        Task { @MainActor in
            do {
                let result = try await pollRepository.deletePoll(withId: pollId)
                onCompletion(result)
            } catch {}
        }
    }
    
    func updatePost(
        text: String?,
        images: [AmityImageData]?,
        videos: [AmityVideoData]?,
        files: [AmityFileData]?,
        metadata: [String: Any]?,
        mentionees: AmityMentioneesBuilder?,
        completion: @escaping (Result<AmityPost, Error>)->()
    ) {
        
        Task { @MainActor in
            guard let post = editedPost else { return }
            
            let videoIds = post.childrenPosts.compactMap({ post in post.getVideoInfo() }).map({$0.fileId})
            let imageIds = post.childrenPosts.compactMap({ post in post.getImageInfo() }).map({$0.fileId})
            let fileIds = post.childrenPosts.compactMap({ post in post.getFileInfo() }).map({$0.fileId})
            
            // if text remain the same, skip updating its value.
            let textToUpdate = (post.data?["text"] as? String != text) ? text : nil
            
            let builder: AmityPostBuilder
            
            if let videos = videos, videos.map({$0.fileId}) != videoIds {
                let videoPostBuilder = AmityVideoPostBuilder()
                if let text = textToUpdate {
                    videoPostBuilder.setText(text)
                }
                videoPostBuilder.setVideos(videos)
                builder = videoPostBuilder
            } else if let images = images, images.map({$0.fileId}) != imageIds {
                let imagePostBuilder = AmityImagePostBuilder()
                if let text = textToUpdate {
                    imagePostBuilder.setText(text)
                }
                imagePostBuilder.setImages(images)
                builder = imagePostBuilder
            } else if let files = files, files.map({$0.fileId}) != fileIds {
                let filePostBuilder = AmityFilePostBuilder()
                if let text = textToUpdate {
                    filePostBuilder.setText(text)
                }
                filePostBuilder.setFiles(files)
                builder = filePostBuilder
            } else {
                let textPostBuilder = AmityTextPostBuilder()
                if let text = textToUpdate {
                    textPostBuilder.setText(text)
                }
                builder = textPostBuilder
            }
            
            Log.add(info: "Updating post with id: \(post.postId)")
            let result: AmityPost
            
            do {
                if let mentionees = mentionees {
                    result = try await postRepository.editPost(withId: post.postId, builder: builder, metadata: metadata, mentionees: mentionees)
                    } else {
                    result = try await postRepository.editPost(withId: post.postId, builder: builder)
                }
                completion(.success(result))
            } catch {
                completion(.failure(error))
            }
            editedPost = nil
        }
    }
    
    func deletePost(at index: Int, hardDelete: Bool, onCompletion: @escaping (_ isSuccess: Bool)->()) {
        guard let post = getPostAtIndex(index: 0) else { return }
        Task { @MainActor in
            let result: Bool
            do {
                if hardDelete {
                    result = try await postRepository.hardDeletePost(withId: post.postId, parentId: nil)
                } else {
                    result = try await postRepository.softDeletePost(withId: post.postId, parentId: nil)
                }
                onCompletion(result)
            } catch {
                onCompletion(false)
            }
        }
    }
    
    func deleteChildImagePost(at index: Int, onCompletion: @escaping (_ isSuccess: Bool)->()) {
        
        guard let post = postCollection?.object(at: index) else { return }
        
        let parentPostId = post.postId
        let childPosts = post.data?["images"] as? [[String: Any]]
        
        let firstChildPost = childPosts?.first
        let childPostId = firstChildPost?["postId"] as? String ?? ""
        
        Log.add(info: "[Delete]: Deleting child post with id: \(childPostId)")
        
        Task { @MainActor in
            do {
                let result = try await postRepository.softDeletePost(withId: childPostId, parentId: parentPostId)
                onCompletion(result)
            } catch {
                onCompletion(false)
            }
        }
    }
    
    func deleteFile(fileId: String) async throws -> Bool {
        return try await fileRepository?.deleteFile(fileId: fileId) ?? false
    }
    
    func addReactionToPost(at index: Int, reaction: String) {
                
        guard let post = getPostAtIndex(index: index) else { return }
        
        Task { @MainActor in
            // If you have already added the same reaction to the post, remove it
            if post.myReactions.contains(reaction) {
                do {
                    let result = try await reactionRepository.removeReaction(reaction, referenceId: post.postId, referenceType: .post)
                    Log.add(info: "[SampleApp]: Reaction Removed \(result)")
                } catch {
                    Log.add(info: "[SampleApp]: Reaction Removed error: \(String(describing: error))")
                }
            } else {
                do {
                    let result = try await reactionRepository.addReaction(reaction, referenceId: post.postId, referenceType: .post)
                    Log.add(info: "[SampleApp]: Reaction Added \(result)")
                } catch {
                    Log.add(info: "[SampleApp]: Reaction Added error: \(String(describing: error))")
                }
            }
        }
        
    }
    
    func sortCurrentFeed(option: AmityPostQuerySortOption) {
        guard feedType != .singlePost else { return }
        self.feedSortOption = option

        switch feedType {
        case .myFeed, .userFeed:
            if let id = userId {
                postCollection = feedRepository.getUserFeed(id, sortBy: option, includeDeleted: includeDeletedPosts)
            } else {
                postCollection = feedRepository.getMyFeedSorted(by: option, includeDeleted: includeDeletedPosts)
            }
        case .community:
            guard let communityId = community?.communityId else { return }
            let communityOption: AmityPostQuerySortOption = option == .firstCreated ? .firstCreated:.lastCreated
            postCollection = feedRepository.getCommunityFeed(withCommunityId: communityId, sortBy: communityOption, includeDeleted: includeDeletedPosts, feedType: postFeedType)
        case .globalFeed, .customPostRankingGlobalFeed, .singlePost:
            break
        }
    }
    
    func sortCurrentFeedCommunity(option: FeedItemDefaultAction) {
        guard feedType != .singlePost else { return }
        switch option {
        case .publishedAndSortFirstCreated:
            self.feedSortOption = .firstCreated
            self.postFeedType = .published
        case .publishedAndSortLastCreated:
            self.feedSortOption = .lastCreated
            self.postFeedType = .published
        case .reviewingAndSortFirstCreated:
            self.feedSortOption = .firstCreated
            self.postFeedType = .reviewing
        case .reviewingAndSortLastCreated:
            self.feedSortOption = .lastCreated
            self.postFeedType = .reviewing
        
        default:
            return
        }
        
        switch feedType {
        case .community:
            guard let communityId = community?.communityId else { return }
            let communityOption: AmityPostQuerySortOption = feedSortOption == .firstCreated ? .firstCreated:.lastCreated
            postCollection = feedRepository.getCommunityFeed(withCommunityId: communityId, sortBy: communityOption, includeDeleted: includeDeletedPosts, feedType: postFeedType)
            sortCommunityHandler?()
        default:
            return
        }
    }
    
    func loadMorePosts() {
        guard feedType != .singlePost, let hasMorePosts = postCollection?.hasNext, hasMorePosts else { return }
        
        postCollection?.nextPage()
    }
    
    func viewCommunityMembership(index: Int) -> String {
        guard let post = self.getPostAtIndex(index: index) else { return "" }
        
        if post.targetType == "community" {
            communityMembership = AmityCommunityMembership(client: AmityManager.shared.client!, andCommunityId: post.targetId)
            if let membership = communityMembership?.getMember(withId: post.postedUserId) {
                var allData = "Community Membership Data"
                allData += "\nDisplay Name: \(membership.displayName)"
                allData += "\nIs Banned: \(membership.isBanned)"
                allData += "\nMetadata: \(membership.metadata)"
                allData += "\nRoles: \(membership.roles)"
                return allData
            } else {
                return "No Membership Data"
            }
        } else {
            return "This post is not community post"
        }
    }
    
    // pending posts
    func approve(post: AmityPost?, completion: ((String) -> Void)?) {
        guard let post = post else { return }
        Task { @MainActor in
            do {
                let result = try await postRepository.approvePost(withId: post.postId)
                completion?("Success! Approved post")
            } catch {
                completion?("Something wrong!, \(error)")
            }
        }
    }
    
    func decline(post: AmityPost?, completion: ((String) -> Void)?) {
        guard let post = post else { return }
        Task { @MainActor in
            do {
                let result = try await postRepository.declinePost(withId: post.postId)
                completion?("Success! Decline post")
            } catch {
                completion?("Something wrong!, \(error)")
            }
        }
    }
}
