//
//  ViewController.swift
//  SampleAppLiveVideo
//
//  Created by Nutchaphon Rewik on 16/9/2563 BE.
//  Copyright © 2563 Amity. All rights reserved.
//

import UIKit
import AmitySDK
import AmityVideoPlayerKit

class PlayerViewController: UIViewController {
    
    var streamToWatch: AmityStream?
    
    @IBOutlet private weak var playerContainer: UIView!

    /// This sample app uses AmityVideoPlayer to play live videos.
    /// The player consumes the stream instance, and works automatically.
    /// At the this moment, the player only provide basic functionality, play and stop.
    ///
    /// For more rich features, it is recommend to use other video players that support RTMP streaming.
    /// You can get the RTMP url, by checking at `.watcherUrl` property.
    ///
    private var player: AmityVideoPlayer?
    
    private var isStarting = true {
        didSet {
            updateActionButton()
        }
    }
    
    private var startedEpoch = Date()
    
    private var videoView: UIView!
    private var loadingIndicatorButton: UIBarButtonItem!
    
    private var timer: Timer?
    @IBOutlet private weak var playTimeView: UIView!
    @IBOutlet private weak var playTimeLabel: UILabel!
    
    // Strong reference, since we add/remove at run-time.
    @IBOutlet private var playButton: UIBarButtonItem!
    
    // Strong reference, since we add/remove at run-time.
    @IBOutlet private var pauseButton: UIBarButtonItem!
    
    deinit {
        stopStream()
        timer?.invalidate()
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        title = streamToWatch?.title ?? "Live"
        playTimeView.isHidden = true
        playTimeView.clipsToBounds = true
        playTimeView.layer.cornerRadius = 8
        timer = Timer.scheduledTimer(withTimeInterval: 1, repeats: true) { [weak self] timer in
            guard let strongSelf = self else {
                return
            }
            let totalSeconds = Int(Date().timeIntervalSince(strongSelf.startedEpoch))
            let minute = totalSeconds / 60
            let seconds = totalSeconds % 60
            strongSelf.playTimeLabel.text = "\(String(format: "%02d", minute)):\(String(format: "%02d", seconds))"
        }
        setpUpStreamView()
        createActivityIndicatorBarButton()
        updateActionButton()
    }
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
        playStream()
    }
    
    private func setpUpStreamView() {
        // Create video view and embed in playerContainer
        videoView = UIView(frame: playerContainer.bounds)
        videoView.autoresizingMask = [.flexibleWidth, .flexibleHeight]
        playerContainer.addSubview(videoView)
        // Tell the player to render the video in this view.
        player = AmityVideoPlayer(client: AppManager.shared.client!)
        player?.renderView = videoView
        player?.delegate = self
    }
    
    private func createActivityIndicatorBarButton(){
        let activityIndicator = UIActivityIndicatorView(style: .medium)
        activityIndicator.startAnimating()
        loadingIndicatorButton = UIBarButtonItem(customView: activityIndicator)
    }
    
    private func updateActionButton() {
        guard let player = player else {
            navigationItem.rightBarButtonItems = []
            return
        }
        if isStarting {
            navigationItem.rightBarButtonItems = [loadingIndicatorButton]
            return
        }
        switch player.mediaState {
        case .stopped:
            navigationItem.rightBarButtonItems = [playButton]
        case .playing:
            navigationItem.rightBarButtonItems = [pauseButton]
        @unknown default:
            assertionFailure("Unhandle player media states")
        }
    }
    
    private func playStream() {
        guard let streamToWatch = streamToWatch else {
            assertionFailure("streamToWatch is not assigned.")
            return
        }
        isStarting = true
        startedEpoch = Date()
        player?.play(streamToWatch, completion: { [weak self] result in
            self?.isStarting = false
            switch result {
            case .failure(let error):
                print("Play video error: \(error.localizedDescription)")
            case .success:
                print("Playing video!!")
            }
        })
    }
    
    private func stopStream() {
        player?.stop()
    }
    
    @IBAction private func pauseButtonDidTouch(_ sender: UIBarButtonItem) {
        stopStream()
    }
    
    @IBAction private func playButtonDidTouch(_ sender: UIBarButtonItem) {
        playStream()
    }

}

extension PlayerViewController: AmityVideoPlayerDelegate {
    
    func amityVideoPlayerMediaStateChanged(_ player: AmityVideoPlayer) {
        updateActionButton()
        switch player.mediaState {
        case .playing:
            playTimeView.isHidden = false
        case .stopped:
            playTimeLabel.text = "00:00"
            playTimeView.isHidden = true
        @unknown default:
            assertionFailure("Unhandled case")
        }
    }
    
}


extension PlayerViewController {
    
    static func make() -> PlayerViewController {
        let storyboard = UIStoryboard(name: "Main", bundle: .main)
        let controller = storyboard.instantiateViewController(identifier: "PlayerViewController") as! PlayerViewController
        return controller
    }
    
}
